import rospy
import enum
import math
import numpy as np
import dynamic_reconfigure.client
from typing import List
from clover import srv
from pymavlink import mavutil
from std_srvs.srv import Trigger
from sensor_msgs.msg import Image
from mavros import mavlink
from mavros_msgs.srv import CommandBool, SetMode, ParamGet, ParamSet, CommandLong
from mavros_msgs.msg import State, ParamValue, Mavlink

class DroneMode(enum.Enum):
    '''
    Drone flight modes.
    '''

    STABILIZED = "STABILIZED"
    GUIDED = "GUIDED"
    ACRO = "ACRO"
    RATTITUDE = "RATTITUDE"
    OFFBOARD = "OFFBOARD"
    MISSION = "AUTO.MISSION"
    RTL = "RTL"

class Waypoint():
    x:float = 0
    y:float = 0
    z:float = 0

    def __init__(self, x:float, y:float, z:float):
        self.x = x
        self.y = y
        self.z = z

class GlobalWaypoint():
    lat:float
    lon: float
    alt: float

    def __init__(self, lat:float, lon:float, alt:float):
        self.lat = lat
        self.lon = lon
        self.alt = alt


class Drone:
    def __init__(self, node_name="flight"):
        rospy.init_node(node_name)

        self.tolerance = 0.1

        self.get_telemetry = rospy.ServiceProxy(
            'get_telemetry', srv.GetTelemetry)
        self._navigate = rospy.ServiceProxy('navigate', srv.Navigate)
        self.navigate_global = rospy.ServiceProxy(
            'navigate_global', srv.NavigateGlobal)
        self.set_position = rospy.ServiceProxy('set_position', srv.SetPosition)
        self.set_velocity = rospy.ServiceProxy('set_velocity', srv.SetVelocity)
        self.set_attitude = rospy.ServiceProxy('set_attitude', srv.SetAttitude)
        self.set_yaw = rospy.ServiceProxy('set_yaw', srv.SetYaw)
        self.set_yaw_rates = rospy.ServiceProxy('set_yaw_rate', srv.SetYawRate)
        self.set_rates = rospy.ServiceProxy('set_rates', srv.SetRates)
        self._land = rospy.ServiceProxy('land', Trigger)
        self._set_mode = rospy.ServiceProxy('mavros/set_mode', SetMode)
        self.send_command = rospy.ServiceProxy(
            'mavros/cmd/command', CommandLong)
        self.release = rospy.ServiceProxy('simple_offboard/release', Trigger)

    def navigate_wait(self, x: float = 0.0, y: float = 0.0, z: float = 0.0, yaw: float = float('nan'), speed: float = 0.5, frame_id: str = 'body', auto_arm: bool = True) -> None:
        '''
        Navigate to a position with wait for completion.
        '''

        telemetry = self.get_telemetry(frame_id='body')

        self._navigate(x=x, y=y, z=z, yaw=yaw, speed=speed,
                       frame_id=frame_id, auto_arm=auto_arm)

        while not rospy.is_shutdown():
            telemetry = self.get_telemetry(frame_id='navigate_target')
            if math.sqrt((telemetry.x)**2 + (telemetry.y)**2 + (telemetry.z)**2) < self.tolerance:
                break
            rospy.sleep(0.1)

    def navigate_global_wait(self, lat: float, lon: float, z: float = 0.0, yaw: float = float('nan'), speed: float = 0.5, frame_id: str = 'body', auto_arm: bool = True) -> None:
        '''
        Navigate to a global position with wait for completion.
        '''

        telemetry = self.get_telemetry(frame_id='body')

        self.navigate_global(lat=lat, lon=lon, z=z, yaw=yaw,
                             frame_id=frame_id, speed=speed, auto_arm=auto_arm)

        while not rospy.is_shutdown():
            telemetry = self.get_telemetry(frame_id='navigate_target')
            if math.sqrt((telemetry.x)**2 + (telemetry.y)**2 + (telemetry.z)**2) < self.tolerance:
                break
            rospy.sleep(0.1)

    def land_wait(self) -> None:
        '''
        Land the drone and wait until it is landed.
        '''

        self._land()
        while self.get_telemetry().armed:
            rospy.sleep(0.2)

    def wait_arrival(self, tolerance: float = 0.1) -> None:
        '''
        Wait until the drone arrives at the target position.
        '''

        while not rospy.is_shutdown():
            telemetry = self.get_telemetry(frame_id='navigate_target')
            if math.sqrt(telemetry.x ** 2 + telemetry.y ** 2 + telemetry.z ** 2) < tolerance:
                break
            rospy.sleep(0.2)

    def get_distance(self, x1: float, y1: float, z1: float, x2: float, y2: float, z2: float) -> float:
        '''
        Calculate the Euclidean distance between two points in 3D space.
        '''

        return math.sqrt((x1 - x2) ** 2 + (y1 - y2) ** 2 + (z1 - z2) ** 2)

    def arm(self, arm: bool = True) -> None:
        '''
        Arm the drone.
        '''

        try:
            arm_service = rospy.ServiceProxy('mavros/cmd/arming', CommandBool)
            arm_service(arm)
        except rospy.ServiceException as e:
            rospy.logerr(f"Service call failed: {e}")

    def is_flipped(self) -> bool:
        '''
        Verifies if drone is flipped.
        '''

        PI_2 = math.pi / 2
        telem = self.get_telemetry()
        flipped = abs(telem.roll) > PI_2 or abs(telem.pitch) > PI_2
        return flipped

    def orbit(self, radius: float = 0.6, speed: float = 0.3) -> None:
        '''
        Make the drone orbit around a point.
        '''

        RADIUS = radius  # m
        SPEED = speed  # rad / s

        start = self.get_telemetry()
        start_stamp = rospy.get_rostime()

        r = rospy.Rate(10)

        while not rospy.is_shutdown():
            angle = (rospy.get_rostime() - start_stamp).to_sec() * SPEED
            x = start.x + math.sin(angle) * RADIUS
            y = start.y + math.cos(angle) * RADIUS
            self.set_position(x=x, y=y, z=start.z)

            r.sleep()

    def send_msg(self, msg: str) -> None:
        '''
        Send a message to the drone.
        '''

        mavlink_pub = rospy.Publisher('mavlink/to', Mavlink, queue_size=1)

        # Sending a HEARTBEAT message:
        msg = mavutil.mavlink.MAVLink_heartbeat_message(
            mavutil.mavlink.MAV_TYPE_GCS, 0, 0, 0, 0, 0)
        msg.pack(mavutil.mavlink.MAVLink('', 2, 1))
        ros_msg = mavlink.convert_to_rosmsg(msg)

        mavlink_pub.publish(ros_msg)

    def set_mode(self, mode: DroneMode) -> None:
        '''
        Set the flight mode of the drone.
        '''

        try:
            self._set_mode(custom_mode=mode)
        except rospy.ServiceException as e:
            rospy.logerr(f"Failed to set mode {mode}: {e}")

    def calibrate_gyro(self):
        '''
        Calibrate the drone's gyro.
        '''

        rospy.loginfo('Calibrate gyro')
        if not self.send_command(command=mavutil.mavlink.MAV_CMD_PREFLIGHT_CALIBRATION, param1=1).success:
            return False

        calibrating = False
        while not rospy.is_shutdown():
            state = rospy.wait_for_message('mavros/state', State)
            if state.system_status == mavutil.mavlink.MAV_STATE_CALIBRATING or state.system_status == mavutil.mavlink.MAV_STATE_UNINIT:
                calibrating = True
            elif calibrating and state.system_status == mavutil.mavlink.MAV_STATE_STANDBY:
                rospy.loginfo('Calibrating finished')
                return True

    def toggle_aruco(self) -> None:
        '''
        Toggle the ArUco marker detection.
        '''

        try:
            aruco_client = dynamic_reconfigure.client.Client('aruco_detection')
            config = aruco_client.get_configuration()
            config['enable'] = not config['enable']
            aruco_client.update_configuration(config)
            rospy.loginfo(
                f"Aruco detection {'enabled' if config['enable'] else 'disabled'}")
        except Exception as e:
            rospy.logerr(f"Failed to toggle ArUco detection: {e}")

    def toggle_optical_flow(self) -> None:
        '''
        Toggle the optical flow detection.
        '''

        try:
            optical_flow_client = dynamic_reconfigure.client.Client(
                'optical_flow')
            config = optical_flow_client.get_configuration()
            config['enable'] = not config['enable']
            optical_flow_client.update_configuration(config)
            rospy.loginfo(
                f"Optical flow {'enabled' if config['enable'] else 'disabled'}")
        except Exception as e:
            rospy.logerr(f"Failed to toggle optical flow: {e}")

    def read_cparam(self, param_name: str) -> float:
        '''
        Read a parameter from flight controller.
        '''

        try:
            param_get = rospy.ServiceProxy('mavros/param/get', ParamGet)
            response = param_get(param_name)
            if response.success:
                return response.value.real
            else:
                rospy.logerr(f"Failed to read parameter {param_name}")
                return float("nan")
        except rospy.ServiceException as e:
            rospy.logerr(f"Service call failed: {e}")
            return float("nan")

    def write_cparam(self, param_name: str, value: float) -> bool:
        '''
        Write a parameter to flight controller.
        '''

        try:
            param_set = rospy.ServiceProxy('mavros/param/set', ParamSet)
            param_value = ParamValue(real=value)
            response = param_set(param_name, param_value)
            if response.success:
                return True
            else:
                rospy.logerr(f"Failed to write parameter {param_name}")
                return False
        except rospy.ServiceException as e:
            rospy.logerr(f"Service call failed: {e}")
            return False

    def haversine_distance(self, lat1, lon1, lat2, lon2, radius=6371000):
        '''
        Calculate the distance between two points on Earth specified by latitude/longitude.
        '''

        # Convert degrees to radians
        phi1, phi2 = math.radians(lat1), math.radians(lat2)
        dphi = math.radians(lat2 - lat1)
        dlambda = math.radians(lon2 - lon1)

        # Haversine formula
        a = math.sin(dphi / 2.0) ** 2 + math.cos(phi1) * \
            math.cos(phi2) * math.sin(dlambda / 2.0) ** 2
        c = 2 * math.atan2(math.sqrt(a), math.sqrt(1 - a))

        return radius * c

    def waypoint_navigate(self, waypoints: List[Waypoint]):
        '''
        Flight through the defined waypoints.
        '''

        for wp in waypoints:
            print(f"note: navigating to waypoint: ({wp.x}, {wp.y}, {wp.z})")
            self.navigate_wait(x=wp.x, y=wp.y, z=wp.z, speed=0.5)
            rospy.sleep(0.5)  # Pause at waypoint

    def global_waypoint_navigate(self, waypoints: List[GlobalWaypoint]):
        '''
        Flight through the defined global waypoints.
        '''

        for i, wp in enumerate(waypoints):
            print(f"Moving to waypoint {i+1}: ({wp.lat}, {wp.lon}, {wp.alt})")
            self.navigate_global_wait(lat=wp.lat, lon=wp.lon, z=wp.alt, speed=0.5)
            rospy.sleep(0.5)  # Pause at waypoint

    def follow(self):
        from .apps.follow import _follow_callback
        rospy.Subscriber('main_camera/image_raw_throttled', Image, _follow_callback, queue_size=1)
        rospy.spin()

    def gesture_control(self):
        from .apps.gesture_control import _gc_callback
        rospy.Subscriber('main_camera/image_raw_throttled', Image, _gc_callback, queue_size=1)
        rospy.spin()






import cv2 as cv
import mediapipe as mp
import numpy as np
import rospy
from sensor_msgs.msg import Image
from cv_bridge import CvBridge
from clover import long_callback
from .handlers.gesture_control import handle_move, handle_fingercount

bridge = CvBridge()
image_pub = rospy.Publisher('~gesture_control/debug', Image, queue_size=1)

mp_hands = mp.solutions.hands
mp_drawing = mp.solutions.drawing_utils
mp_styles = mp.solutions.drawing_styles
hands = mp_hands.Hands(max_num_hands=2, model_complexity=0, min_detection_confidence=0.5, min_tracking_confidence=0.5)


@long_callback
def _gc_callback(data):
    frame = bridge.imgmsg_to_cv2(data, 'bgr8')
    blank = np.zeros((500, 500, 3), dtype="uint8")

    frame = cv.flip(frame, 1)
    frame.flags.writeable = False
    frame = cv.cvtColor(frame, cv.COLOR_BGR2RGB)
    results = hands.process(frame)
    frame.flags.writeable = True
    frame = cv.cvtColor(frame, cv.COLOR_RGB2BGR)

    h,w,_ = frame.shape

    points = []
    if results.multi_hand_landmarks:
        for hand_landmark in results.multi_hand_landmarks:
            mp_drawing.draw_landmarks(blank, hand_landmark, mp_hands.HAND_CONNECTIONS, mp_styles.get_default_hand_landmarks_style(), mp_styles.get_default_hand_connections_style())

            for index, cord in enumerate(hand_landmark.landmark):
                cx, cy = int(cord.x * w), int(cord.y * h)
                points.append((cx, cy))

            count = 0
            if hand_landmark:
                count = handle_fingercount(points)

            cv.putText(blank, str(count), (20, 70), cv.FONT_HERSHEY_COMPLEX, 1, (255, 255, 255), 2)

            handle_move(count)

    image_pub.publish(bridge.cv2_to_imgmsg(frame, 'bgr8'))



from ...drone import Drone

drone = Drone()

def handle_move(count):
    if count == 1:
        drone.navigate_wait(x=0, y=0, z=0.1, frame_id='body', auto_arm=True)
    elif count == 2:
        drone.navigate_wait(x=0, y=0, z=-0.1, frame_id='body')
    elif count == 3:
        drone.navigate_wait(x=0.1, y=0, z=0, frame_id='body')
    elif count == 4:
        drone.navigate_wait(x=-0.1, y=0, z=0, frame_id='body')
    else:
        drone.navigate_wait(x=0, y=0, z=0, frame_id='body')
    return

def handle_fingercount(points):
    # RIGHT HAND PALM
    THUMB = 4
    fingers = [8, 12, 16, 20]
    count = 0

    if points[THUMB][0] < points[THUMB - 2][0]:
        count += 1

    for x in fingers:
        if points[x][1] < points[x - 2][1]:
            count += 1

    return count



from ...drone import Drone

drone = Drone()

def follow_handle_move(target_size: float):
    if 250 <= target_size <= 340:
        pass
        # drone.navigate_wait(x=0, y=0, z=0.0, frame_id='body', speed=0.5, auto_arm=True)
    elif target_size < 250:
        drone.navigate_wait(x=0.1, y=0.0, z=0.0, frame_id='body', speed=0.5, auto_arm=True)
    elif target_size > 340:
        drone.navigate_wait(x=-0.1, y=0, z=0, frame_id='body', speed=0.5, auto_arm=True)

def centralize_in_target(center: float, landmark0, w):
    pos = landmark0.x * w
    if center - 100 <= pos <= center + 100:
        drone.set_yaw(yaw=radians(0), frame_id='body')
    elif pos > center + 100:
        drone.set_yaw(yaw=radians(5), frame_id='body')
    elif pos < center - 100:
        drone.set_yaw(yaw=radians(-5), frame_id='body')




import cv2 as cv
from mediapipe import solutions
from mediapipe.framework.formats import landmark_pb2
import numpy as np

def resize(frame, scale=1):
    width = int(frame.shape[1] * scale)
    height = int(frame.shape[0] * scale)
    dimensions = (width, height)

    return cv.resize(frame, dimensions, interpolation=cv.INTER_AREA)

def changeRes(capture, width, height):
    capture.set(3, width)
    capture.set(4, height)

def draw_landmarks_on_image(rgb_image, detection_result):
  pose_landmarks_list = detection_result.pose_landmarks
  annotated_image = np.copy(rgb_image)

  for idx in range(len(pose_landmarks_list)):
    pose_landmarks = pose_landmarks_list[idx]

    pose_landmarks_proto = landmark_pb2.NormalizedLandmarkList()
    pose_landmarks_proto.landmark.extend([
      landmark_pb2.NormalizedLandmark(x=landmark.x, y=landmark.y, z=landmark.z) for landmark in pose_landmarks
    ])
    solutions.drawing_utils.draw_landmarks(
      annotated_image,
      pose_landmarks_proto,
      solutions.pose.POSE_CONNECTIONS,
      solutions.drawing_styles.get_default_pose_landmarks_style())
  return annotated_image




import rospy
import piexif
import cv2
import os
from datetime import datetime
from clover import srv
from cv_bridge import CvBridge
from sensor_msgs.msg import Image
from fractions import Fraction

class Camera():
    def __init__(self):
        self.bridge = CvBridge()

    def retrieve_cv_frame(self):
        '''
        Retrieve a single frame.
        '''

        return self.bridge.imgmsg_to_cv2(rospy.wait_for_message('main_camera/image_raw_throttled', Image), 'bgr8')

    def save_image(self, path:str):
        timestamp = datetime.now().strftime('%Y-%m-%d_%H-%M-%S')
        filename = os.path.join(path, timestamp + '.jpg')

        frame = self.retrieve_cv_frame()
        cv2.imwrite(filename, frame)

        telemetry = rospy.ServiceProxy('get_telemetry', srv.GetTelemetry)()

        lat = telemetry.lat
        lon = telemetry.lon
        alt = telemetry.alt

        def to_deg(value):
            frac = Fraction(value).limit_denominator()
            return ((abs(frac.numerator), abs(frac.denominator)),)

        exif_dict = {
            "GPS": {
                piexif.GPSIFD.GPSLatitudeRef: "N" if lat >= 0 else "S",
                piexif.GPSIFD.GPSLatitude: to_deg(lat),
                piexif.GPSIFD.GPSLongitudeRef: "E" if lon >= 0 else "W",
                piexif.GPSIFD.GPSLongitude: to_deg(lon),
                piexif.GPSIFD.GPSAltitude: (int(alt), 1),
                piexif.GPSIFD.GPSAltitudeRef: 0,
            }
        }

        exif_bytes = piexif.dump(exif_dict)
        piexif.insert(exif_bytes, filename)




import time
from math import trunc
import cv2 as cv
import mediapipe as mp
from ..cv.utils import draw_landmarks_on_image
from sensor_msgs.msg import Image
from cv_bridge import CvBridge
from clover import long_callback
from .handlers.follow import follow_handle_move

import rospy
import os

BASE_DIR = os.path.dirname(os.path.abspath(__file__))
model_path= os.path.join(BASE_DIR, '..', 'cv', 'models', 'pose_landmarker_full.task')
BaseOptions = mp.tasks.BaseOptions
PoseLandmarkerOptions = mp.tasks.vision.PoseLandmarkerOptions
VisionRunningMode = mp.tasks.vision.RunningMode
detection_results = None
bridge = CvBridge()

def result_callback(result, output_image, timestamp_ms):
    detection_results = result

options = PoseLandmarkerOptions(
    base_options=BaseOptions(model_asset_path=model_path),
    running_mode=VisionRunningMode.LIVE_STREAM,
    num_poses=1,
    min_pose_detection_confidence=0.65, result_callback=result_callback)

landmarker = mp.tasks.vision.PoseLandmarker.create_from_options(options)
image_pub = rospy.Publisher('~follow/debug', Image, queue_size=1)

def set_target_size(result, h):
    try:
        pose_landmarks = result.pose_landmarks[0]
        y0 = pose_landmarks[0].y
        y1 = pose_landmarks[23].y
    except (IndexError, TypeError):
        return None
    return (y1 - y0) * h

@long_callback
def _follow_callback(data):
    frame = bridge.imgmsg_to_cv2(data,'bgr8')

    frame = cv.flip(frame, 1)

    frame_timestamp_ms = int(time.time() * 1000)
    mp_image = mp.Image(image_format=mp.ImageFormat.SRGB, data=frame)
    landmarker.detect_async(mp_image, frame_timestamp_ms)

    if detection_results is not None:
        h, w, _ = frame.shape
        center = trunc(w / 2)
        target_size = set_target_size(detection_results, h)

        if target_size is not None:
            landmark0 = detection_results.pose_landmarks[0][0]
            centralize_in_target(center, landmark0, w)
            follow_handle_move(target_size)

        annotated_frame= draw_landmarks_on_image(frame, detection_results)
        frame = cv.cvtColor(annotated_frame, cv.COLOR_RGB2BGR)

    image_pub.publish(bridge.cv2_to_imgmsg(frame, 'bgr8'))



import pigpio
import time

class Servo():
    def __init__(self, gpio: int):
        self.gpio = gpio
        self.pi = pigpio.pi()

        if not self.pi.connected:
            raise Exception("Pi not connected")

        self.pi.set_mode(self.gpio, pigpio.OUTPUT)

    def pwm_neutral(self, sleep=0.45):
        '''
        Set servo to neutral position.
        '''

        try:
            self.pi.set_servo_pulsewidth(self.gpio, 1000)
            time.sleep(sleep)
            self.pi.write(self.gpio, 0)
        except KeyboardInterrupt:
            self.pi.write(self.gpio, 0)

    def pwm_high(self, sleep=0.45):
        '''
        Set servo to the highest secure pulsewidth.
        '''

        try:
            self.pi.set_servo_pulsewidth(self.gpio, 2000)
            time.sleep(sleep)
            self.pi.write(self.gpio, 0)
        except KeyboardInterrupt:
            self.pi.write(self.gpio, 0)

    def pwm_low(self, sleep=0.45):
        '''
        Set servo to the lowest secure pulsewidth.
        '''

        try:
            self.pi.set_servo_pulsewidth(self.gpio, 500)
            time.sleep(sleep)
            self.pi.write(self.gpio, 0)
        except KeyboardInterrupt:
            self.pi.write(self.gpio, 0)

    def set_pulsewidth(self, sleep=0.45, pulsewidth=1500):
        '''
        Set servo to any pulsewidth. Not recommended in most cases.
        '''

        try:
            self.pi.set_servo_pulsewidth(self.gpio, pulsewidth)
            time.sleep(sleep)
            self.pi.write(self.gpio, 0)
        except KeyboardInterrupt:
            self.pi.write(self.gpio, 0)





# Information: https://clover.coex.tech/programming

import rospy
from abc import ABC, abstractmethod
from .drone import Drone, DroneMode
from .camera import Camera
from .servo import Servo
from typing import Union

class Task(ABC):
    '''
    An abstract class to write mission.
    '''

    RTL_ALTITUDE = 10

    def __init__(self, gpio: Union[int, None] = None) -> None:
        if gpio != None:
            self.servo = Servo(gpio)

        self.drone = Drone()
        self.camera = Camera()

    @abstractmethod
    def mission(self):
        raise Exception("Need implementation.")

    def run(self):
        '''
        A secure method to run a mission. Useful in most cases.
        '''

        try:
            self.mission()

        except KeyboardInterrupt:
            print("warning: aborting task")

        except Exception as e:
            print(f"ERROR: {e}")

        finally:
            print('note: landing')
            self.drone.land_wait()

    def return_to_launch_confim(self):
        '''
        Use if you need to confirm the return.
        '''
        confirm = input("Return to launch position? (Y/n): ").strip().lower()
        if confirm == 'y':
            self.drone.set_mode(DroneMode.RTL)

    def change_servo_pin(self, gpio:int):
        '''
        Change the servo gpio.
        '''

        self.servo.gpio = gpio

